from datetime import datetime, timedelta

from crispy_forms.layout import Layout, Row, Column
from django import forms
from django.core.exceptions import ValidationError
from django.utils import timezone

from django.utils.translation import gettext_lazy as _

from medux import notifications
from medux.common.api.interfaces import ILoginViewExtension, ILoginFormExtension
from medux.common.bootstrap import RadioSelectButtonsGroup
from .bootstrap import TimeSelector
from medux.common.tools import round_down_time

from medux.employees.models import Employee
from medux.plugins.timetracker.models import TimeEntry

COME_GO_CHOICES = (
    ("come", _("Coming")),
    ("go", _("Going")),
)


def get_rounded_time():
    """helper function to get current time, rounded down to the next unit."""
    return round_down_time(timezone.localtime(), 15).strftime("%H:%M")


class TimetrackerLoginFormExtension(ILoginFormExtension):
    """Provides simplified start/end time fields at the login form."""

    class Media:
        js = ("timetracker/js/timeselector.js",)

    field_names = ["log_time", "come_go"]

    def alter_fields(self, fields: dict):
        """add log_time and come/go fields to Login dialog"""
        fields["log_time"] = forms.TimeField(
            label=_("Login time"),
            required=False,
            initial=get_rounded_time,
        )
        fields["come_go"] = forms.ChoiceField(
            label=False, required=False, choices=COME_GO_CHOICES
        )

    def alter_layout(self, layout: Layout):
        """insert a timeselector and come/go fields into layout"""
        fields: list = layout.fields[0].fields
        fields.insert(
            2,
            Row(
                Column(
                    TimeSelector("log_time"),
                    RadioSelectButtonsGroup("come_go"),
                ),
            ),
        )

    def clean(self, cleaned_data) -> dict:
        if not cleaned_data["come_go"] in ["come", "go"]:
            return cleaned_data
        employee = Employee.objects.get(username=cleaned_data["username"])
        open_time_entry = TimeEntry.objects.filter(
            employee=employee,
            work_schedule=employee.active_work_schedule,
            end=None,
        )
        if cleaned_data["come_go"] == "come":
            if open_time_entry.exists():
                raise ValidationError(
                    _(
                        f"You already seem to have an active time entry at "
                        f"{open_time_entry.last().start.strftime('%d.%m.%Y %H:%m')}. "
                        f"Please finish this first.",
                    ),
                    code="invalid",
                )
        elif cleaned_data["come_go"] == "go":
            if open_time_entry.exists():
                if len(open_time_entry) != 1:
                    raise ValidationError(
                        _(
                            "There are more than one unfinished time entries. Please contact your administrator."
                        ),
                        code="invalid",
                    )
                else:
                    open_time_entry = open_time_entry.first()
                    today = datetime.date(datetime.today())
                    timestamp = timezone.make_aware(
                        datetime.combine(today, cleaned_data["log_time"])
                    )
                    if timestamp < open_time_entry.start:
                        raise ValidationError(
                            _(
                                f"End time ({timestamp}) cannot be before start time ({open_time_entry})."
                            ),
                            code="invalid",
                        )
        return cleaned_data


class TimetrackerLoginViewExtension(ILoginViewExtension):
    def form_valid(self, form, response) -> None:
        """After successfully logging in, save some timetracker data."""
        if not form.cleaned_data["come_go"]:
            return
        today = datetime.date(datetime.today())
        try:
            employee = Employee.objects.get(username=form.cleaned_data["username"])
            timestamp = timezone.make_aware(
                datetime.combine(today, form.cleaned_data["log_time"])
            )
            open_time_entry = TimeEntry.objects.filter(
                employee=employee, work_schedule=employee.active_work_schedule, end=None
            )
            if form.cleaned_data["come_go"] == "come":
                # TODO: make sure employee has an active_work_schedule!
                TimeEntry.objects.create(employee=employee, start=timestamp)
            elif form.cleaned_data["come_go"] == "go":
                if open_time_entry.exists():
                    open_time_entry = open_time_entry[0]
                    if open_time_entry.start < timestamp:
                        open_time_entry.end = timestamp
                        open_time_entry.save()
                    else:
                        notifications.error(
                            _("End time cannot be before start time"),
                            recipient=employee,
                        )  # TODO

        except Employee.DoesNotExist:
            pass



from django.urls import reverse, reverse_lazy

from medux.common.api.interfaces import IMenuItem
from django.utils.translation import gettext_lazy as _

from medux.plugins.timetracker.models import TimeEntry


# def has_unfinished_timeentry(request):
#     """returns True if user is employee and has unfinished time slots"""
#     if request.user.is_anonymous or not request.user.is_employee:
#         return False
#     return TimeEntry.objects.filter(employee=request.user.employee, end=None).exists()


class TimetrackerDashboard(IMenuItem):
    menu = "views"
    title = _("Time tracker")
    icon = "calendar-range"
    url = reverse("timetracker:overview")


class Overview(IMenuItem):
    menu = "timetracker"
    title = _("Overview")
    icon = "house"
    separator = True
    url = reverse("timetracker:overview")


class TimeEntryList(IMenuItem):
    menu = "timetracker"
    title = _("Time entries")
    icon = "clock"
    url = reverse("timetracker:timeentry-list")


class Holidays(IMenuItem):
    menu = "timetracker"
    title = _("Holidays")
    icon = "airplane"
    url = reverse("timetracker:current-holidays")


class LogoutWithTimeTracking(IMenuItem):
    menu = "user"
    title = _("Logout with time tracking")
    url = "#"
    # url = reverse_lazy("timetracker:logout")
    weight = 90
    icon = "box-arrow-right"
    attrs = {"hx-get": reverse_lazy("timetracker:logout"), "hx-target": "#dialog"}


class Administration(IMenuItem):
    menu = "timetracker"
    title = _("Administration")
    icon = "gear"
    url = reverse("timetracker:admin")
    required_permissions = "timetracker.administer_holiday"



from django.test.signals import setting_changed
from gdaps.conf import PluginSettings

# This is a default conf file for a GDAPS plugin.
# You can use settings anywhere in your plugin using this syntax:
#
#     from .conf import timetracker_settings
#     # or:
#     # from /home/christian/Projekte/timetracker_dev/medux/plugins.timetracker.conf import timetracker_settings
#
#     foo = timetracker_settings.FOO_SETTING
#
# This way you can use custom (plugin-default) settings, that can be overridden globally if needed.


# required parameter.
NAMESPACE = 'TIMETRACKER'

# Optional defaults. Leave empty if not needed.
DEFAULTS = {
    # 'MY_SETTING': 'somevalue',
    # 'FOO_PATH': '/home/christian/Projekte/timetracker_dev/medux/plugins.timetracker.models.FooModel',
    # 'BAR': [
    #     'baz',
    #     'buh',
    # ],
}

# Optional list of settings keys that are allowed to be in 'string import' notation. Leave empty if not needed.
IMPORT_STRINGS = (
    # 'FOO_PATH',
)

# Optional list of settings that have been removed. Leave empty if not needed.
REMOVED_SETTINGS = ()


timetracker_settings = PluginSettings(NAMESPACE, DEFAULTS, IMPORT_STRINGS)


def reload_timetracker_settings(*args, **kwargs):
    setting = kwargs['setting']
    if setting == 'TIMETRACKER':
        timetracker_settings.reload()


setting_changed.connect(reload_timetracker_settings)



from datetime import timedelta

from django.contrib.auth.signals import user_logged_in
from django.db.models import QuerySet
from django.utils.timezone import now

# from medux import notifications
from medux.plugins.timetracker.models import TimeEntry


def notify_employee_after_login_for_open_timeentries(**kwargs):
    user = kwargs["user"]
    if hasattr(user, "employee"):
        employee = user.employee
        if employee.active_work_schedule:
            # search for started, but unfinished timeentries of the current user
            current_timeentries: QuerySet = TimeEntry.objects.filter(
                work_schedule=employee.active_work_schedule,
                end=None,
                start__lt=now() + timedelta(minutes=15),
                # work_schedule__working_hours__week_day=today().weekday(),
            )
            # there are unfinished timeentries, notify user
            # notifications.success(
            #     kwargs["user"],
            #     "There are unfinished time entries.",  # TODO link to Timetracker
            #     dismissible=True,
            # )


user_logged_in.connect(notify_employee_after_login_for_open_timeentries)



from crispy_forms.layout import Field
from crispy_forms.utils import TEMPLATE_PACK


class TimeSelector(Field):
    """Crispy Widget that provides a time input, the current time with seconds,
    and two buttons that increase/decrease the time by a unit, per default 15min.
    """

    template = "bootstrap5/layout/timeselector.html"

    def __init__(self, *args, show_current_time=False, units=15, **kwargs):
        super().__init__(*args, **kwargs)
        self.show_current_time = show_current_time
        self.units = units

    def render(
        self,
        form,
        form_style,
        context,
        template_pack=TEMPLATE_PACK,
        extra_context=None,
        **kwargs
    ):
        if extra_context is None:
            extra_context = {}
        extra_context["show_current_time"] = self.show_current_time
        extra_context["units"] = self.units
        extra_context["step"] = self.units * 60
        return super().render(
            form, form_style, context, template_pack, extra_context, **kwargs
        )



import datetime

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout
from django import forms
from django.forms import ModelForm
from django.core.exceptions import ValidationError
from django.utils import timezone
from django.utils.formats import time_format
from django.utils.translation import gettext as _

from medux.common.tools import round_down_time, round_up_time
from medux.plugins.timetracker.bootstrap import TimeSelector
from medux.plugins.timetracker.models import TimeEntry


class TimeEntryForm(ModelForm):
    class Meta:
        model = TimeEntry
        fields = ["date", "start_time", "end_time", "comment"]

    class Media:
        js = ("timetracker/js/timeselector.js",)

    date = forms.DateField()
    start_time = forms.TimeField()
    end_time = forms.TimeField(required=False)

    def __init__(self, **kwargs):
        self.employee = kwargs.pop("employee")
        super().__init__(**kwargs)
        if self.instance.id:
            self.fields["date"].initial = timezone.localdate(self.instance.start)
            self.fields["start_time"].initial = time_format(
                timezone.localtime(self.instance.start)
            )
            self.fields["end_time"].initial = time_format(
                timezone.localtime(self.instance.end)
            )
        else:
            self.fields["date"].initial = timezone.localtime().date()

            self.fields["start_time"].initial = time_format(
                round_down_time(timezone.localtime(self.instance.start), 15)
            )
        self.helper = FormHelper()
        self.helper.form_tag = False
        self.helper.layout = Layout(
            "date",
            TimeSelector("start_time"),
            TimeSelector("end_time"),
            "comment",
        )

    def clean(self):
        assert self.employee.active_work_schedule
        self.cleaned_data["employee"] = self.employee
        if not "end_time" in self.cleaned_data and TimeEntry.objects.filter(
            employee=self.employee, end=None
        ):
            self.add_error(
                "end_time",
                ValidationError(
                    _(
                        "There can only be one open time entry. Please close the open one first."
                    ),
                    code="invalid",
                ),
            )
        return super().clean()

    def clean_end_time(self):
        end_time = self.cleaned_data.get("end_time", None)
        start_time = self.cleaned_data.get("start_time", None)
        if end_time:
            if end_time <= start_time:
                self.add_error(
                    "end_time",
                    ValidationError(
                        _("End time cannot be before start time."), code="invalid"
                    ),
                )
        return end_time

    def save(self, commit=True):
        instance: TimeEntry = self.instance
        instance.start = timezone.make_aware(
            datetime.datetime.combine(
                self.cleaned_data["date"], self.cleaned_data["start_time"]
            )
        )
        if self.cleaned_data["end_time"]:
            instance.end = timezone.make_aware(
                datetime.datetime.combine(
                    self.cleaned_data["date"], self.cleaned_data["end_time"]
                )
            )
        instance.employee = self.employee
        # if new TimeEntry was created, set work_schedule to default
        if not instance.id:
            instance.work_schedule = self.employee.active_work_schedule
        return super().save(commit)


class LogoutWithTimeTrackingForm(TimeEntryForm):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.fields["end_time"].required = True
        self.fields["end_time"].initial = time_format(
            round_up_time(timezone.localtime(), 15)
        )



from django.contrib.auth import get_user_model
from django.urls import path, include
from rest_framework import routers, serializers, viewsets

from medux.plugins.timetracker.models import TimeEntry
from . import views

User = get_user_model()

app_name = "timetracker"


class TimeEntrySerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = TimeEntry
        fields = "__all__"


class TimeEntryViewSet(viewsets.ModelViewSet):
    queryset = TimeEntry.objects.all()
    serializer_class = TimeEntrySerializer


# Serializers define the API representation.
class UserSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = User
        fields = ["url", "username", "email", "is_staff"]


# ViewSets define the view behavior.
class UserViewSet(viewsets.ModelViewSet):
    queryset = User.objects.all()
    serializer_class = UserSerializer


# Routers provide an easy way of automatically determining the URL conf.
# router = routers.DefaultRouter()
# router.register(r"users", UserViewSet)
# router.register(r"timeentries", TimeEntryViewSet)

# Wire up our API using automatic URL routing.
# Additionally, we include login URLs for the browsable API.
urlpatterns = [
    # path("api/", include(router.urls)),
    path("api-auth/", include("rest_framework.urls", namespace="rest_framework")),
    path("overview/", views.OverView.as_view(), name="overview"),
    path(
        "overview/<int:year>/<int:month>/",
        views.OverView.as_view(),
        name="overview-month",
    ),
    path("holidays/", views.HolidaysView.as_view(), name="current-holidays"),
    path("holidays/<int:year>/", views.HolidaysView.as_view(), name="holidays"),
    path(
        "toggle-holiday/<int:year>/<int:month>/<int:day>/",
        views.ToggleHolidayView.as_view(),
        name="toggle-holiday",
    ),
    path(
        "set-holiday-status/<pk>/<str:status>",
        views.HolidaySetStatusView.as_view(),
        name="set-holiday-status",
    ),
    path("admin/", views.AdminView.as_view(), name="admin"),
    path(
        "pending-holidays/",
        views.PendingHolidayRequestsView.as_view(),
        name="pending-holiday-requests",
    ),
    path(
        "delete-holiday/<pk>/", views.DeleteHolidayView.as_view(), name="delete-holiday"
    ),
    # TimeEntries
    path("timeentry/", views.TimeEntryListView.as_view(), name="timeentry-list"),
    path("timeentry/add/", views.TimeEntryCreateView.as_view(), name="timeentry-add"),
    path(
        "timeentry/<pk>/change/",
        views.TimeEntryUpdateView.as_view(),
        name="timeentry-update",
    ),
    path(
        "timeentry/<pk>/delete/",
        views.TimeEntryDeleteView.as_view(),
        name="timeentry-delete",
    ),
    # WorkingTimeRange
    path(
        "work_schedule/",
        views.CurrentWorkScheduleView.as_view(),
        name="current-work-schedule",
    ),
    # generic
    path("logout/", views.LogoutWithTimeTrackingView.as_view(), name="logout"),
]



import datetime
from calendar import (
    LocaleHTMLCalendar,
    month_name,
    January,
    different_locale,
)
import holidays
from dateutil.utils import today
from django.db.models import QuerySet
from django.template import Engine, Context

from medux.common.tools import django_locale_to_country_code
from medux.employees.models import Employee
from medux.plugins.timetracker.models import Holiday, HolidayStatus


def get_holiday_context_from_day(
    employee: Employee,
    thedate: datetime.date,
    queryset: QuerySet[Holiday],
    locale: tuple,
) -> dict:
    """Helper function that creates context for one holiday.

    Returns:
        context: a dict with usable context for rendering one day:
            day: the date of the day to render
            object: the Holiday object of the employee
            type: the `HolidayType` of the holiday
            status: the `HolidayStatus` of the holiday
    """

    context = {}
    context["day"] = thedate
    try:
        obj = queryset.get(employee_id=employee.id, day=thedate)
        context["object"] = obj
        context["status"] = obj.status
        context["type"] = obj.type
        context["approved"] = obj.status == HolidayStatus.APPROVED
        context["declined"] = obj.status == HolidayStatus.DECLINED
        context["consumed"] = (
            obj.status == HolidayStatus.APPROVED and thedate < today().date()
        )
    except Holiday.DoesNotExist:
        obj = False
    context["collisions"] = queryset.filter(day=thedate).order_by("employee_id")
    country_code = django_locale_to_country_code(locale)
    is_public = thedate in holidays.country_holidays(country_code)
    context["public"] = is_public

    is_weekend = thedate.weekday() in [5, 6]
    context["weekend"] = is_weekend

    is_editable = (
        not is_weekend
        and not is_public
        and (not obj or obj.status == HolidayStatus.NEW)
    )
    context["editable"] = is_editable
    return context


class HolidayCalendar(LocaleHTMLCalendar):
    queryset = None

    def __init__(self, employee: Employee, queryset=None, locale=None, firstweekday=0):
        self.queryset = queryset or Holiday.objects.none()
        self.employee = employee
        super().__init__(firstweekday, locale)

    def formatday(self, thedate: datetime.date):
        """Return a day as a table cell."""
        context = get_holiday_context_from_day(
            employee=self.employee,
            thedate=thedate,
            queryset=self.queryset,
            locale=self.locale,
        )
        return (
            Engine.get_default()
            .get_template("timetracker/day.html")
            .render(Context(context))
        )

    def formatmonthname(self, theyear, themonth):
        """Return a month name as a table header."""
        with different_locale(self.locale):
            return f"<th class='{self.cssclass_month_head}'>{month_name[themonth]}</th>"

    def formatmonth(self, theyear, themonth):
        """
        Return a month as a table row.
        """
        v = []
        a = v.append
        a('<tr class="month">')
        a(self.formatmonthname(theyear, themonth))
        for week in self.monthdatescalendar(theyear, themonth):
            for day in week:
                if day.month == themonth:
                    a(self.formatday(day))

        a("</tr>")
        return "".join(v)

    def formatyear(self, employee, theyear):
        """
        Return a formatted year as a table with months as rows.
        """
        self.queryset = self.queryset.filter(day__year=theyear)
        v = []
        a = v.append
        a('<div class="table-responsive">')
        a(f"<table class='table table-borderless {self.cssclass_year}'>")
        for month in range(January, January + 12):
            a(self.formatmonth(theyear, month))
        a("</table>")
        a("</div>")
        return "".join(v)



import locale
import calendar

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import logout
from django.http import HttpResponse
from django.utils.timezone import datetime, timedelta

from django.contrib.auth.mixins import PermissionRequiredMixin
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.views.generic import (
    TemplateView,
    ListView,
    DetailView,
    CreateView,
    UpdateView,
    FormView,
)
from django.utils.translation import gettext_lazy as _

from medux.common.api.interfaces import ModalFormViewMixin
from medux.common.htmx.mixins import HtmxResponseMixin, HtmxDeleteView
from medux.common.tools import monthdelta
from medux.employees.models import WorkSchedule
from medux.plugins.timetracker.forms import TimeEntryForm, LogoutWithTimeTrackingForm
from medux.plugins.timetracker.models import Holiday, HolidayStatus, TimeEntry
from medux.plugins.timetracker import calendars
from medux.core.models import User
from medux import notifications
from medux.common.api.http import HttpResponseEmpty, HttpResponseHXRedirect


class CalendarContextMixin:
    """Adds calendar data to the context.
    Arguments:
        year: the year to create context data from. If not present, the current year is taken
        month: the month to create context data from. If not present, the current month is taken
    """

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        now = timezone.now()
        year = kwargs.get("year", now.year)
        month = kwargs.get("month", now.month)
        first_day = datetime(year, month, 1)
        last_day = monthdelta(first_day, 1) - timedelta(days=1)
        prev_month = monthdelta(first_day, -1)
        next_month = monthdelta(first_day, 1)
        current_month = now.month
        current_year = now.year
        context.update(
            {
                "year": year,
                "month": month,
                "first_day": first_day,
                "last_day": last_day,
                "month_label": _(first_day.strftime("%B")),
                "prev_month": prev_month,
                "next_month": next_month,
                "prev_year": year - 1,
                "next_year": year + 1,
                "current_month": current_month,
                "current_year": current_year,
            }
        )
        return context


class HolidaysView(PermissionRequiredMixin, CalendarContextMixin, TemplateView):
    template_name = "timetracker/holidays.html"
    permission_required = "timetracker.view_holiday"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        queryset = Holiday.objects.filter(
            employee__tenant=self.request.user.tenant
        ).order_by("employee_id")
        # FIXME: add firstweekday, locale as parameters
        # noinspection PyTypeChecker
        cal = calendars.HolidayCalendar(
            employee=self.request.user,
            queryset=queryset,
            firstweekday=1,
        )
        context["calendar"] = mark_safe(
            cal.formatyear(self.request.user, context["year"])
        )
        return context


class OverView(PermissionRequiredMixin, CalendarContextMixin, TemplateView):
    template_name = "timetracker/overview.html"

    def has_permission(self):
        user = self.request.user
        return user.is_authenticated and user.has_perm("timetracker.view_timeentry")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        # Generate a calendar
        now = timezone.now()
        month = context["month"] or now.month
        year = context["year"] or now.year

        cal = calendar.Calendar(firstweekday=0).itermonthdates(year=year, month=month)
        # context["calendar"] = cal.monthdatescalendar(context["year"], context["month"])
        return context


class ToggleHolidayView(HtmxResponseMixin, PermissionRequiredMixin, TemplateView):
    template_name = "timetracker/day.html"
    object: Holiday | None = None
    queryset = None

    def get_queryset(self, **kwargs):
        if self.queryset:
            return self.queryset
        self.queryset = Holiday.objects.filter(
            employee__tenant=self.request.user.tenant
        ).order_by("employee_id")
        return self.queryset

    def has_permission(self):
        user: User = self.request.user
        return user.is_employee and user.has_perm("timetracker.change_holiday")

    def get_context_data(self, year, month, day, **kwargs):
        context = super().get_context_data(**kwargs)
        thedate = timezone.datetime(year, month, day)
        queryset = self.get_queryset()
        try:
            self.object = queryset.get(employee_id=self.request.user.id, day=thedate)
            context["object"] = self.object
        except Holiday.DoesNotExist:
            pass
        context.update(
            calendars.get_holiday_context_from_day(
                employee=self.request.user.employee,
                thedate=thedate,
                queryset=queryset,
                locale=locale.getdefaultlocale(),
            )
        )
        context["day"] = thedate
        return context

    def post(self, request, **kwargs):
        context = self.get_context_data(**self.kwargs)
        if self.object:
            if self.object.status == HolidayStatus.APPROVED:
                notifications.warning(
                    self.request.user, _("Holiday was already approved.")
                )
                return HttpResponseEmpty()
            elif self.object.status == HolidayStatus.NEW:
                pass
            self.object.delete()
            self.object = None
            del context["object"]
        else:
            # there is no holiday marked on this day
            self.object = Holiday.objects.create(
                employee_id=self.request.user.id, day=context["day"], type_id=1
            )
            context["object"] = self.object

        return self.render_to_response(context)


class AdminView(PermissionRequiredMixin, CalendarContextMixin, TemplateView):
    template_name = "timetracker/admin.html"

    def has_permission(self):
        return self.request.user.has_perm("timetracker.administer_holiday")


class PendingHolidayRequestsView(PermissionRequiredMixin, ListView):
    """Shows all pending holiday requests from employees in a list."""

    template_name = "timetracker/pending_holiday_list.html"
    permission_required = "timetracker.administer_holiday"
    model = Holiday

    def get_queryset(self):
        return Holiday.objects.filter(
            employee__tenant=self.request.user.tenant, status=HolidayStatus.NEW
        ).order_by("day")


class HolidaySetStatusView(HtmxResponseMixin, PermissionRequiredMixin, DetailView):
    """Sets status (approved, declined,...) of the given Holiday."""

    def get_queryset(self):
        return Holiday.objects.filter(
            employee__tenant=self.request.user.tenant, status=HolidayStatus.NEW
        ).order_by("day")

    def has_permission(self):
        user: User = self.request.user
        return (
            user.has_perm("timetracker.change_holiday")
            and self.get_object().employee.tenant == user.tenant
        )

    def post(self, request, status: str, **kwargs):
        obj = self.get_object()
        obj.status = HolidayStatus(status)
        obj.save()
        return HttpResponse()


class DeleteHolidayView(PermissionRequiredMixin, HtmxDeleteView):
    model = Holiday

    def has_permission(self):
        user = self.request.user
        # only allow deleting own holidays, or if user has "administer" perm
        return user.has_perm("timetracker.delete_holiday") and (
            user.id == self.get_object().employee.id
            or user.has_perm("timetracker.administer_holiday")
        )

    def get_queryset(self):
        return Holiday.objects.filter(
            employee__tenant=self.request.user.tenant,
            status__in=[
                HolidayStatus.APPROVED,
                HolidayStatus.NEW,
                HolidayStatus.DECLINED,
            ],
        ).order_by("day")


class TimeEntryListView(PermissionRequiredMixin, ListView):
    model = TimeEntry
    permission_required = "timetracker.view_timeentry"
    template_name = "timetracker/timeentry_list.html"

    def get_queryset(self):
        user = self.request.user
        if not user.is_employee:
            messages.error(
                self.request, _("Time entries are only available for employees.")
            )
            return TimeEntry.objects.none()
        employee = user.employee
        if not employee.active_work_schedule:
            messages.warning(
                self.request,
                _("You haven't created a work schedule yet. Please create one."),
                extra_tags="alert-dismissible",
            )
        return TimeEntry.objects.filter(
            employee=employee, work_schedule=employee.active_work_schedule
        ).order_by("-start")


class TimeEntryCreateView(ModalFormViewMixin, CreateView):
    model = TimeEntry
    form_class = TimeEntryForm
    modal_title = _("Add new time entry")
    success_event = "medux:timeentry:changed"

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs.update({"employee": self.request.user.employee})
        return kwargs


class TimeEntryUpdateView(ModalFormViewMixin, PermissionRequiredMixin, UpdateView):
    model = TimeEntry
    form_class = TimeEntryForm
    modal_title = _("Edit time entry")
    success_event = "medux:timeentry:changed"

    def has_permission(self):
        # TODO: maybe require timetracker.change|add_timeentry permission here?
        return self.request.user.is_employee

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs.update(
            {"employee": self.request.user.employee, "instance": self.get_object()}
        )
        return kwargs


class CurrentWorkScheduleView(HtmxResponseMixin, PermissionRequiredMixin, DetailView):
    model = WorkSchedule
    template_name = "timetracker/work_schedule_detail.html"

    def has_permission(self):
        user = self.request.user
        return user.is_authenticated and user.is_employee

    def get_queryset(self):
        return self.model.objects.filter(employee=self.request.user.employee)

    def get_object(self, queryset=None):
        if not self.kwargs.get("pk", None):
            return WorkSchedule.get_active(self.request.user.employee)
        return super().get_object(queryset)


class TimeEntryDeleteView(PermissionRequiredMixin, HtmxDeleteView):
    model = TimeEntry
    success_event = "medux:timeentry:changed"

    def has_permission(self):
        user = self.request.user
        return (
            user.is_employee
            and user.has_perm("timetracker.delete_timeentry")
            and self.get_object().employee == user.employee
        )


class LogoutWithTimeTrackingView(TimeEntryUpdateView):
    """A modal form view that allows finishing a time entry before logging out."""

    template_name = "timetracker/logout_with_timetracking_form.html"
    modal_title = _("Logout with time tracking")
    model = TimeEntry

    def get_queryset(self):
        return self.model.objects.filter(employee=self.request.user.employee)

    def get_object(self, queryset=None):
        queryset = queryset or self.get_queryset()
        try:
            # first, try to get open time entry
            return queryset.get(end=None)
        except TimeEntry.DoesNotExist:
            # if no open time entry exists, create new one.
            work_schedule: WorkSchedule = (
                self.request.user.employee.active_work_schedule
            )
            if not work_schedule:
                return None
            weekday = timezone.localdate().weekday()
            start = work_schedule.working_hours.filter(weekday=weekday)
            if not start.exists():
                # no WTR defined for today's weekday
                return None

            return TimeEntry(
                employee=self.request.user.employee,
                work_schedule=work_schedule,
                start=start.first(),
            )

    def get_form(self, form_class=None):
        employee = self.request.user.employee
        try:
            instance = TimeEntry.objects.get(employee=employee, end=None)
        except TimeEntry.DoesNotExist:
            instance = None
        form = LogoutWithTimeTrackingForm(
            data=self.request.POST or None, instance=instance, employee=employee
        )
        return form

    def form_valid(self, form):
        # save the time entry
        super().form_valid(form)
        # and log user out.
        logout(self.request)
        return HttpResponseHXRedirect(redirect_to=settings.LOGIN_URL)

    def form_invalid(self, form):
        print(
            "form.errors:",
            form.errors,
            "\nform._errors:",
            form._errors,
            "\nself.non_field_errors:",
            form.non_field_errors(),
        )
        return super().form_invalid(form)



from datetime import timedelta, datetime

import django
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import TextChoices
from django.utils import timezone
from django.utils.formats import time_format, date_format
from django.utils.translation import gettext_lazy as _

from medux.employees.models import Employee, WorkSchedule, WorkingTimeRange


def _roundTime(dt: datetime = None, dateDelta: timedelta = timedelta(minutes=1)):
    """Round a datetime object to a multiple of a timedelta

    @param dt: datetime.datetime object, default now.
    dateDelta : timedelta object, we round to a multiple of this, default 1 minute.
    Author: Thierry Husson 2012 - Use it as you want but don't blame me.
            Stijn Nevens 2014 - Changed to use only datetime objects as variables
            Christian González 2020 - adjustments and small improvements
    """
    roundTo = dateDelta.total_seconds()

    if dt is None:
        dt = django.utils.timezone.now()
    # Make sure dt and datetime.min have the same timezone
    tzmin = dt.min.replace(tzinfo=dt.tzinfo)

    seconds = (dt - tzmin).seconds
    rounding = (seconds + roundTo / 2) // roundTo * roundTo
    return dt + timedelta(0, rounding - seconds, -dt.microsecond)


# class ScheduleEntry(models.Model):
#     """Entry of a roster which is a plan for working hours."""
#
#     schedule = models.ForeignKey(WorkSchedule, on_delete=models.CASCADE)
#     day = models.CharField(max_length=2, choices=DaysOfWeek.choices)
#     start = models.TimeField()
#     end = models.TimeField()
#     comment = models.CharField(max_length=255, blank=True)
#
#     def __str__(self):
#         return f"{self.day} {self.start} - {self.end}"


class TimeEntryManager(models.Manager):
    def create(self, **kwargs):
        """Set work_schedule to employee's active one if not provided."""
        if not "work_schedule" in kwargs:
            kwargs["work_schedule"] = kwargs["employee"].active_work_schedule
        return super().create(**kwargs)


class TimeEntry(models.Model):
    """A time entry for a specific block of time, of one person."""

    # TODO: Schedules must not overlap
    # TODO: time entries of one user must not overlap
    # validate time entries:
    #   TODO: start ! >= end

    objects = TimeEntryManager()

    employee = models.ForeignKey(Employee, on_delete=models.CASCADE)
    work_schedule = models.ForeignKey(WorkSchedule, on_delete=models.PROTECT)
    start = models.DateTimeField(default=timezone.now)
    end = models.DateTimeField(blank=True, null=True)
    comment = models.CharField(
        max_length=255, blank=True, help_text=_("Optional comment for this time entry")
    )
    cached_delta = models.DurationField(_("Delta"), blank=True, null=True)

    def duration(self) -> timedelta:
        """Returns the duration of the time entry as timedelta"""
        if self.end:
            return self.end - self.start
        assert timezone.is_aware(self.start)  # DEBUG
        return timezone.localtime() - self.start

    def duration_h(self) -> float:
        return self.duration().seconds / 60 / 60

    def __str__(self):
        if self.end:
            end = time_format(timezone.localtime(self.end).time())
        else:
            end = _("now")
        return f"{self.employee}: {time_format(timezone.localtime(self.start).time())} - {end}"


    def save(self, **kwargs):
        # FIXME: rounding is dependent on current tenant, hence must be done in view.
        #     if PreferencesRegistry.get["ROUND"]:
        #         round = settings.TIMETRACKER["ROUND"]
        #         self.start = _roundTime(self.start, timedelta(minutes=round))
        #         if self.end:
        #             self.end = _roundTime(self.end, timedelta(minutes=round))
        #

        self.delta(recalculate=True)
        super().save(**kwargs)

    def wtr(self) -> WorkingTimeRange | None:
        """returns WorkingTimeRange that matches this TimeEntry."""

        # get a list ot WorkingTimeRanges for this day
        working_hours = self.work_schedule.working_hours.filter(
            weekday=self.start.weekday()
        )
        # if no working_hours are defined yet, just return empty WTR
        if not working_hours:
            return None

        thetime = self.start
        match = (
            working_hours.filter(start_time__lte=thetime, end_time__gt=thetime)
            .order_by("-start_time")
            .first()
        )
        if not match:
            next_wtr = (
                working_hours.filter(start_time__gt=thetime)
                .order_by("start_time")
                .first()
            )
            prev_wtr = (
                working_hours.filter(end_time__lte=thetime)
                .order_by("-end_time")
                .first()
            )
            if next_wtr and prev_wtr:
                if (thetime - prev_wtr.end).total_seconds() < (
                    next_wtr.start - thetime
                ).total_seconds():
                    match = prev_wtr
                else:
                    match = next_wtr
            elif next_wtr:
                # there is only a next block, no previous
                match = next_wtr
            else:
                # there is only a previous block, and no next one.
                match = prev_wtr
        return match

    def delta(self, recalculate=False) -> timedelta:
        """returns calculated delta between current duration and scheduled duration.

        Note:
            the WTR (incl. cached_delta) is not saved in this method. This must be done manually.
        """
        if recalculate or not self.cached_delta:
            wtr = self.wtr()
            if not wtr:
                return timedelta()
            self.cached_delta = self.duration() - wtr.duration()
        return self.cached_delta

    def delta_h(self) -> float:
        delta = self.delta()
        if delta.days < 0:
            sign = -1
        else:
            sign = 1

        total_seconds = abs(delta.seconds)
        return total_seconds / 60 / 60 * sign
        # hours = total_seconds // 3600
        # minutes = (total_seconds % 3600) // 60
        # return f"{sign}{hours}:{minutes}"


class HolidayType(models.Model):
    """Type of holiday: normal, special,..."""

    name = models.CharField(max_length=255)
    # color = ColorField(blank=True, null=True)

    def __str__(self):
        return self.name


class HolidayStatus(TextChoices):
    NEW = ("new", _("New"))
    APPROVED = ("approved", _("Approved"))
    # CONSUMED = APPROVED and date < today
    DECLINED = ("declined", _("Declined"))


class Holiday(models.Model):
    """Represents one holiday for one user."""

    class Meta:
        unique_together = ["day", "employee"]
        permissions = (
            (
                "administer_holiday",
                _("Can administer holidays"),
            ),
        )

    day = models.DateField()
    employee = models.ForeignKey(Employee, on_delete=models.CASCADE)
    type = models.ForeignKey(HolidayType, on_delete=models.PROTECT)
    status = models.CharField(
        max_length=25, choices=HolidayStatus.choices, default=HolidayStatus.NEW
    )

    def __str__(self):
        # confirmed = "✅" if self.status == HolidayStatus.CONFIRMED else "❌"
        return f"{self.employee}: {date_format(self.day,'SHORT_DATE_FORMAT')} ({self.type}) [{self.status}]"



from django.utils.translation import gettext_lazy as _

from gdaps.api import require_app

# from gdaps.frontend.conf import frontend_settings
from . import __version__
from medux.common.api import MeduxPluginAppConfig


class TimetrackerConfig(MeduxPluginAppConfig):
    """A GDAPS Django app plugin.

    It needs a special parameter named ``PluginMeta``. It is the key for GDAPS
    to recognize this app as a GDAPS plugin.
    ``PluginMeta`` must point to a class that implements certain attributes
    and methods.
    """

    name = "medux.plugins.timetracker"

    groups_permissions = {
        "Users": {
            "timetracker.TimeEntry": ["view", "add", "change", "delete"],
            "timetracker.Holiday": ["view", "add", "change", "delete"],
        },
        "Holiday admins": {"timetracker.Holiday": ["administer"]},
    }

    class PluginMeta:
        """This configuration is the introspection data for plugins."""

        # the plugin machine "name" is taken from the AppConfig, so no name here
        verbose_name = _("Timetracker")
        author = "Christian González"
        author_email = "office@nerdocs.at"
        vendor = "Nerdocs"
        description = _("Timetracking for employees")
        category = _("Base")
        visible = True
        version = __version__
        # compatibility = "medux.core>=2.3.0"

    def ready(self):
        # This function is called after the app and all models are loaded.
        #
        # You can do some initialization here, but beware: it should rather
        # return fast, as it is called at each Django start, even on
        # management commands (makemigrations/migrate etc.).
        #
        # Avoid interacting with the database especially 'save' operations,
        # if you don't *really* have to."""

        try:
            from . import signals
        except ImportError:
            pass

        require_app(self, "medux")
